package net.davidc.egp.snaxe;

import java.awt.Point;
import java.util.*;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @author David Croft (david@davidc.net)
 * @version $Id$
 */
public class GameBoard
{
  private static final Logger log = Logger.getLogger(GameBoard.class.getName());

  private static final int RABBIT_SCORE = 20;

  private static final float RABBIT_SPAWN_CHANCE = 0.2f;
  private static final int RABBIT_PIECE_MIN_LIFETIME_MOVES = 15;
  private static final int RABBIT_PIECE_MAX_LIFETIME_MOVES = 45;

  private static final float INVULNERABILITY_SPAWN_CHANCE = 0.01f;
  private static final int INVULNERABILITY_PIECE_MIN_LIFETIME_MOVES = 15;
  private static final int INVULNERABILITY_PIECE_MAX_LIFETIME_MOVES = 30;
  private static final int INVULNERABILITY_DURATION_TIME = 20 * 1000;

  private static final float SPEED_SPAWN_CHANCE = 0.01f;
  private static final int SPEED_PIECE_MIN_LIFETIME_MOVES = 15;
  private static final int SPEED_PIECE_MAX_LIFETIME_MOVES = 25;
  private static final int SPEED_DURATION_MOVES = 15;

  private static final float SKULL_SPAWN_CHANCE = 0.05f;
  private static final int SKULL_PIECE_MIN_LIFETIME_MOVES = 5;
  private static final int SKULL_PIECE_MAX_LIFETIME_MOVES = 20;
  private static final int SKULL_DEAD_SNAKE_LIFETIME_MOVES = 15;

  int width;
  int height;
  private PlayState playState;
  Set<Snake> snakes;
  Set<Piece> pieces;

  private final Random random = new Random();

  private int minRabbits;
  private int maxRabbits;

  public GameBoard(int width, int height, PlayState playState)
  {
    this.width = width;
    this.height = height;
    this.playState = playState;
    snakes = new HashSet<Snake>();
    pieces = new HashSet<Piece>();
  }

  public void initRabbits()
  {
    minRabbits = (int) Math.ceil(snakes.size() / 4.0);
    maxRabbits = (int) Math.ceil(snakes.size() * 0.75);

    populateRabbits();
//    addPiece(PieceType.SPEED, 10, 10, 1000, 1000);
  }

  long lastTime = 0;

  public void move(long curTime)
  {
    long diffTime = curTime - lastTime;
    lastTime = curTime;

    // Move all snakes
    for (Snake snake : snakes) {
      moveSnake(snake);

      // Reduce invulnerability
      if (snake.invulnerabilityTimeRemaining > 0) {
        snake.invulnerabilityTimeRemaining -= diffTime;
        if (snake.invulnerabilityTimeRemaining <= 0) {
          snake.invulnerabilityTimeRemaining = 0;
          invulnerabilityEnded(snake);
        }
      }

      if (snake.speedMovesRemaining > 0) {
        snake.speedMovesRemaining--;
      }
    }

    // if a snake is now off-grid, it dies.
    for (Snake snake : snakes) {
      checkSnakeOffGrid(snake);
    }

    // if two snakes occupy the same position, both die
    for (Snake snake1 : snakes) {
      checkSnakeHeadbuttSnake(snake1);
    }

    // if a snake occupies the position of someone else's body or tail, they die
    for (Snake snake1 : snakes) {
      SnakePart partToCheck = snake1.head;
      checkSnakeOnSnake(snake1, partToCheck);
    }

    // if a snake occupies the position of a board piece, they die or gain the powerup
    for (Snake snake : snakes) {
      checkSnakeOnPiece(snake, snake.head.x, snake.head.y);
    }

    // Kill all dying snakes
    killDeadSnakes();

    // if a power-up occupies the space of another powerup (e.g. two skulls, or a skull on top of a rabbit), keep
    // only one (currently randomish)
    killStackedPieces();

    // Now decay old powerups
    decayPieces();

    // Now add new powerups
    populateRabbits();
    spawnSpecials();
  }

  // Half-move, for speedy snakes only
  public void halfMove()
  {
    // Move all snakes
    for (Snake snake : snakes) {
      if (snake.speedMovesRemaining > 0) {
        moveSnake(snake);
      }
    }

    // if a snake is now off-grid, it dies.
    for (Snake snake : snakes) {
      if (snake.speedMovesRemaining > 0) {
        checkSnakeOffGrid(snake);

        // if two snakes occupy the same position, both die
        checkSnakeHeadbuttSnake(snake);

        // if a snake occupies the position of someone else's body or tail, they die
        SnakePart partToCheck = snake.head;
        checkSnakeOnSnake(snake, partToCheck);

        // if a snake occupies the position of a board piece, they die or gain the powerup
        checkSnakeOnPiece(snake, snake.head.x, snake.head.y);
      }
    }

    // Kill all dying snakes
    killDeadSnakes();
  }

  private void moveSnake(Snake snake)
  {
    snake.move();
  }

  private void checkSnakeOffGrid(Snake snake)
  {
    if (snake.head.x < 0 || snake.head.x >= width || snake.head.y < 0 || snake.head.y >= height) {
      log.log(Level.FINE, "Snake " + snake + " went off-board and died.");
      snake.dead = true;
    }
  }

  private void checkSnakeHeadbuttSnake(Snake snake1)
  {
    if (snake1.dead) return;
    if (snake1.invulnerabilityTimeRemaining > 0) return;

    for (Snake snake2 : snakes) {
      if (snake1 == snake2) continue;
      if (snake2.invulnerabilityTimeRemaining > 0) continue; // invulnerable

      if (snake1.head.x == snake2.head.x && snake1.head.y == snake2.head.y) {
        log.log(Level.FINE, "Snake " + snake1 + " headbutted " + snake2 + ". Both died.");
        snake1.dead = true;
        snake2.dead = true;
      }
    }
  }

  private void checkSnakeOnSnake(Snake snake1, SnakePart partToCheck)
  {
    int x = partToCheck.x;
    int y = partToCheck.y;

    if (snake1.dead) return;
    if (snake1.invulnerabilityTimeRemaining > 0) return;

    nextSnake2:
    for (Snake snake2 : snakes) {
      if (snake2.invulnerabilityTimeRemaining > 0) continue; // invulnerable

      SnakePart part = snake2.head;
//        if (snake1 == snake2) part = part.next; // allow to touch their own head! - now partToCheck != part below
      while (part != null) {
        if (partToCheck != part && x == part.x && y == part.y) {
          log.log(Level.FINE, "Snake " + snake1 + " hit the tail of " + snake2 + " and died.");
          snake1.dead = true;
          break nextSnake2;
        }
        part = part.next;
      }
    }
  }

  private void checkSnakeOnPiece(Snake snake, int x, int y)
  {
    if (snake.dead) return;

    nextPiece:
    for (Piece piece : pieces) {
      if (x == piece.x && y == piece.y) {

        switch (piece.type) {
          case RABBIT:
            snake.score += RABBIT_SCORE;
            snake.grow++;
            playState.addPowerUpRiseEvent(piece.x, piece.y, "+" + RABBIT_SCORE);
            log.log(Level.FINE, "Snake " + snake + " ate a rabbit, score now " + snake.score);
            break;
          case SKULL:
            if (snake.invulnerabilityTimeRemaining > 0) continue nextPiece; // invulnerable
            log.log(Level.FINE, "Snake " + snake + " was killed by a skull.");
            snake.dead = true;
            break;
          case INVULNERABILITY:
            snake.invulnerabilityTimeRemaining = INVULNERABILITY_DURATION_TIME;
            playState.addPowerUpRiseEvent(piece.x, piece.y, "INVULNERABLE!");
            log.log(Level.FINE, "Snake " + snake + " got invulnerability");
            break;
          case SPEED:
            snake.speedMovesRemaining = SPEED_DURATION_MOVES;
            playState.addPowerUpRiseEvent(piece.x, piece.y, "+SPEED");
            log.log(Level.FINE, "Snake " + snake + " got speed");
            break;
        }

        pieces.remove(piece);
        return; // should never be two pieces on top of each other, so this is ok
      }
    }
  }

  private void killDeadSnakes()
  {
    boolean done = false;
    killSnakes:
    while (!done) {
      for (Snake snake : snakes) {
        if (snake.dead) {
          // Replace all pieces by skulls
          log.log(Level.FINE, "Snake " + snake + " is replaced by skulls.");
          SnakePart part = snake.head.next; // don't place a skull on the head itself, cos that's the position of whatever killed them!
          while (part != null) {
            addPiece(PieceType.SKULL, part.x, part.y, SKULL_DEAD_SNAKE_LIFETIME_MOVES, SKULL_DEAD_SNAKE_LIFETIME_MOVES);
            part = part.next;
          }

          // Remove snake
          snakes.remove(snake);
          // Modified our iterators; must restart
          continue killSnakes;
        }
      }
      done = true;
    }
  }

  private void killStackedPieces()
  {
    boolean done;

    done = false;
    restartPiece:
    while (!done) {
      for (Piece piece1 : pieces) {
        for (Piece piece2 : pieces) {
          if (piece1 == piece2) continue;
          if (piece1.x == piece2.x && piece1.y == piece2.y) {
            log.log(Level.FINE, "Piece " + piece1 + " is on top of " + piece2 + ", the latter is removed.");
            pieces.remove(piece2);
            // Modified our iterators; must restart
            continue restartPiece;
          }
        }
      }
      done = true;
    }
  }

  private void decayPieces()
  {
    List<Piece> piecesToRemove = new ArrayList<Piece>();

    for (Piece piece : pieces) {
      piece.lifetimeRemaining--;
      if (piece.lifetimeRemaining <= 0) {
        piecesToRemove.add(piece);
        log.log(Level.FINE, "Piece " + piece + " decayed");
      }
    }

    for (Piece piece : piecesToRemove) {
      pieces.remove(piece);
    }
  }

  private void invulnerabilityEnded(Snake snake)
  {
    //Snake just exited invulnerability. Check that none of its parts cross each other, another snake, or a skull

    // if two snakes occupy the same position, both die
    checkSnakeHeadbuttSnake(snake);

    SnakePart part = snake.head;
    while (part != null) {
      // if this part occupies the position of someone else's or their own body or tail, they die
      checkSnakeOnSnake(snake, part);

      // if this part occupies the position of a board piece, they die or gain the powerup
      checkSnakeOnPiece(snake, part.x, part.y);

      part = part.next;
    }
  }

  private void populateRabbits()
  {
    int numRabbits = countPieces(PieceType.RABBIT);

    if (numRabbits >= maxRabbits) return;

    while (numRabbits < minRabbits) {
      addRandomRabbit();
      numRabbits++;
    }

    // Chance of adding an extra rabbit

    if (random.nextFloat() < RABBIT_SPAWN_CHANCE) {
      addRandomRabbit();
    }
  }

  private void spawnSpecials()
  {
    int numSpecials = countPieces(PieceType.INVULNERABILITY) + countPieces(PieceType.SPEED);
    if (numSpecials == 0) {
      if (random.nextFloat() < INVULNERABILITY_SPAWN_CHANCE) {
        addRandomInvulnerability();
      }
      else if (random.nextFloat() < SPEED_SPAWN_CHANCE) {
        addRandomSpeed();
      }
    }

    if (random.nextFloat() < SKULL_SPAWN_CHANCE) {
      addRandomSkull();
    }
  }

  private int countPieces(PieceType type)
  {
    int numPiecesOfType = 0;
    for (Piece piece : pieces) {
      if (piece.type == type) {
        numPiecesOfType++;
      }
    }
    return numPiecesOfType;
  }

  private void addRandomRabbit()
  {
    Point p = getRandomUnoccupiedPoint();
    if (p == null) return;

    log.log(Level.FINE, "Add rabbit at " + p.x + "," + p.y);
    addPiece(PieceType.RABBIT, p.x, p.y, RABBIT_PIECE_MIN_LIFETIME_MOVES, RABBIT_PIECE_MAX_LIFETIME_MOVES);
  }

  private void addRandomInvulnerability()
  {
    Point p = getRandomUnoccupiedPoint();
    if (p == null) return;

    log.log(Level.FINE, "Add invulnerability at " + p.x + "," + p.y);
    addPiece(PieceType.INVULNERABILITY, p.x, p.y, INVULNERABILITY_PIECE_MIN_LIFETIME_MOVES, INVULNERABILITY_PIECE_MAX_LIFETIME_MOVES);
  }

  private void addRandomSpeed()
  {
    Point p = getRandomUnoccupiedPoint();
    if (p == null) return;

    log.log(Level.FINE, "Add speed at " + p.x + "," + p.y);
    addPiece(PieceType.SPEED, p.x, p.y, SPEED_PIECE_MIN_LIFETIME_MOVES, SPEED_PIECE_MAX_LIFETIME_MOVES);
  }

  private void addRandomSkull()
  {
    Point p = getRandomUnoccupiedPoint();
    if (p == null) return;

    log.log(Level.FINE, "Add skull at " + p.x + "," + p.y);
    addPiece(PieceType.SKULL, p.x, p.y, SKULL_PIECE_MIN_LIFETIME_MOVES, SKULL_PIECE_MAX_LIFETIME_MOVES);
  }

  private Point getRandomUnoccupiedPoint()
  {
    Point p = new Point();

    newPoint:
    for (int i = 0; i < 50; ++i) {
      p.x = random.nextInt(width);
      p.y = random.nextInt(height);

      // Make sure there's not an existing powerup or snake part there
      for (Piece piece : pieces) {
        if (piece.x == p.x && piece.y == p.y) continue newPoint;
      }

      for (Snake snake : snakes) {
        SnakePart part = snake.head;
        while (part != null) {
          if (part.x == p.x && part.y == p.y) continue newPoint;
          part = part.next;

        }
      }

      return p;
    }

    log.log(Level.WARNING, "Couldn't find a place for new powerup");
    return null;
  }


  private void addPiece(PieceType type, int x, int y, int minLifetime, int maxLifetime)
  {
    int lifetime = minLifetime + random.nextInt(maxLifetime + 1 - minLifetime);
    pieces.add(new Piece(type, x, y, lifetime));
  }
}
