package net.davidc.egp.snaxe.networking;

import java.io.*;
import java.net.Socket;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This class acts as a screen client to Edward Mccaughan's swarm server. A screen client receives messages
 * from all controller clients and implements the game logic.
 *
 * @author David Croft (david@davidc.net)
 * @version $Id: SwarmScreenClient.java 26 2011-09-18 16:24:33Z david $
 * @see <a href="https://github.com/edwardmccaughan/pubbage-undecimal/tree/master/snaxe">https://github.com/edwardmccaughan/pubbage-undecimal/tree/master/snaxe</a>
 */
public class SwarmScreenClient
{
  private static final Logger log = Logger.getLogger(SwarmScreenClient.class.getName());

  private static final String NETWORK_CHARSET = "UTF-8";

  private static final String PROTOCOL_DELIMITER = "_";
  private static final String PROTOCOL_NEWLINE = "\r\n";

  private static final String INCOMING_MESSAGE_PLAYER = "m";
  private static final String INCOMING_MESSAGE_NEW = "new";
  private static final String INCOMING_MESSAGE_QUIT = "quit";

  private static final String OUTGOING_MESSAGE_PLAYER = "p" + PROTOCOL_DELIMITER;
  private static final String OUTGOING_MESSAGE_BROADCAST = "p" + PROTOCOL_DELIMITER + "a" + PROTOCOL_DELIMITER;

  private Socket socket;

  private List<SwarmListener> swarmListeners;

  private BufferedReader inputReader;
  private Writer outputWriter;
  private InputStream inputStream;

  public SwarmScreenClient()
  {
    swarmListeners = new LinkedList<SwarmListener>();
  }

  /**
   * Connect to the swarm server on the given host and port.
   *
   * @param host The host to connect to.
   * @param port The port to connect to.
   * @throws IOException if connection failed
   */
  public void connect(String host, int port) throws IOException
  {
    socket = new Socket(host, port);
//    socketChannel = SocketChannel.open();
//    socketChannel.connect(new InetSocketAddress(host, port));
//    socketChannel.configureBlocking(false);

    inputStream = socket.getInputStream();
//    inputStream = new SocketChannelInputStream(socket);
    inputReader = new BufferedReader(new InputStreamReader(inputStream, NETWORK_CHARSET));
    OutputStream outputStream = socket.getOutputStream();

    outputWriter = new BufferedWriter(new OutputStreamWriter(outputStream, NETWORK_CHARSET));
    log.log(Level.INFO, "Connected to swarm server " + host + " on port " + port);
  }

  /**
   * Add a listener to be informed of events from the swarm server.
   * @param listener the listener to add
   */
  public void addSwarmListener(SwarmListener listener)
  {
    swarmListeners.add(listener);
  }

  /**
   * Remove a previously-added swarm listener.
   * @param listener the listener to add
   */
  public void removeSwarmListener(SwarmListener listener)
  {
    swarmListeners.remove(listener);
  }

  /**
   * Send a message to the given player.<p/>
   * Protocol: p_&lt;playerId&gt;_&lt;message&gt;\r\n
   *
   * @param playerId The player to send to
   * @param message  The message to send
   * @throws IOException if sending failed
   */
  public void sendMessage(int playerId, String message) throws IOException
  {
    StringBuilder sb = new StringBuilder();
    sb.append(OUTGOING_MESSAGE_PLAYER);
    sb.append(playerId);
    sb.append(PROTOCOL_DELIMITER);
    sb.append(message);
    sb.append(PROTOCOL_NEWLINE);

    log.log(Level.FINEST, "Sending: " + sb.toString());

    outputWriter.write(sb.toString());
    outputWriter.flush();
  }

  /**
   * Send a message to all player.<p/>
   * Protocol: p_a_&lt;message&gt;\r\n
   *
   * @param message The message to send
   * @throws IOException if sending failed
   */
  public void broadcastMessage(String message) throws IOException
  {
    StringBuilder sb = new StringBuilder();
    sb.append(OUTGOING_MESSAGE_BROADCAST);
    sb.append(message);
    sb.append(PROTOCOL_NEWLINE);

    log.log(Level.FINEST, "Sending: " + sb.toString());

    outputWriter.write(sb.toString());
    outputWriter.flush();
  }

  /**
   * Close the connection to the swarm server.
   *
   * @throws IOException if the closing failed
   */
  public void close() throws IOException
  {
    socket.close();
  }

  /**
   * Polls the swarm server for new messages, dispatching these to registered listeners as appropriate.
   *
   * @throws IOException if polling failed
   */
  public void poll() throws IOException
  {
    if (socket.isClosed()) {
      throw new IOException("Server socket is closed");
    }
    if (!socket.isConnected()) {
      throw new IOException("Server socket is not connected");
    }

    // Incoming messages delimited by _, format messageType_params (e.g. p_playerNumber_message)
    inputStream.available();

    while (inputReader.ready()) {
      String message = inputReader.readLine();

      if (message == null) {
        throw new IOException("END OF STREAM"); // TODO NOT WORKING
      }

      log.log(Level.FINEST, "Received: " + message);

      String[] bits = message.split(PROTOCOL_DELIMITER, 3);

      if (bits[0].equals(INCOMING_MESSAGE_PLAYER)) {
        int playerId = Integer.parseInt(bits[1]);
        for (SwarmListener swarmListener : swarmListeners) {
          swarmListener.messageReceived(playerId, bits[2]);
        }
      }
      else if (bits[0].equals(INCOMING_MESSAGE_NEW)) {
        int playerId = Integer.parseInt(bits[1]);
        for (SwarmListener swarmListener : swarmListeners) {
          swarmListener.playerJoined(playerId);
        }
      }
      else if (bits[0].equals(INCOMING_MESSAGE_QUIT)) {
        int playerId = Integer.parseInt(bits[1]);
        for (SwarmListener swarmListener : swarmListeners) {
          swarmListener.playerQuit(playerId);
        }
      }
      else {
        log.log(Level.WARNING, "Unrecognised message from server: " + message);
      }
    }
  }
}
