package net.davidc.egp.snaxe.states;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.davidc.egp.common.resources.ResourcesException;
import net.davidc.egp.common.slick2d.SlickGraphicUtil;
import net.davidc.egp.common.slick2d.resources.SlickResources;
import net.davidc.egp.snaxe.Player;
import net.davidc.egp.snaxe.SnaxeGame;
import net.davidc.egp.snaxe.networking.Protocol;
import org.newdawn.slick.*;
import org.newdawn.slick.state.BasicGameState;
import org.newdawn.slick.state.StateBasedGame;

/**
 * @author David Croft (david@davidc.net)
 * @version $Id: InsertCoinState.java 26 2011-09-18 16:24:33Z david $
 */
public class InsertCoinState extends BasicGameState
{
  private static final Logger log = Logger.getLogger(InsertCoinState.class.getName());

  public static final int STATE_ID = 1;
  private static final int QUEUED_PLAYERS_PER_LINE = 5;

  private SlickResources resources;

  private static enum SubState
  {
    WAITING, COUNTDOWN, LAUNCHED
  }

  private SubState subState;

  private long curTime;
  private long countdownStartTime;
  private int countdownLength;

  private int lastNotifiedCountdown; // Last countdown notified to players

  // Colours to flicker through for the title text
  private static final Color[] TITLE_COLORS = new Color[]{Color.white, Color.red, Color.green, Color.blue, Color.yellow, Color.cyan, Color.magenta};

  // Countdown length, indexed by number of players queued
  private static final int COUNTDOWN_LENGTHS[] = new int[]{
          -1, // 0 players
          -1, // 1 player
          60, // 2 players
          55, // 3 players
          50, // 4 players
          40, // 5 players
          30, // 6 players
          20, // 7 players
          0, // 8 players
  };

  // If F12 was pressed to skip the countdown
  private boolean forceStart;

  @Override
  public int getID()
  {
    return STATE_ID;
  }

  public void init(GameContainer container, StateBasedGame game) throws SlickException
  {
    resources = new SlickResources(((SnaxeGame) game).getGlobalResources());
    try {
      resources.loadFromClasspath("/resources.xml", "insertcoin");
    }
    catch (ResourcesException e) {
      throw new SlickException("Unable to load resources", e);
    }
  }

  @Override
  public void enter(GameContainer container, StateBasedGame game) throws SlickException
  {
    subState = SubState.WAITING;
    curTime = 0;
    lastNotifiedCountdown = -1;
    forceStart = false;
  }

  public void render(GameContainer container, StateBasedGame game, Graphics g) throws SlickException
  {
    AngelCodeFont font = resources.getFont("FONT_NOKIA48");

    int x = container.getWidth() / 2;
    int y = container.getHeight() / 2 - (font.getLineHeight() * 4);

    g.setColor(TITLE_COLORS[((int) ((curTime / 50) % TITLE_COLORS.length))]);

    SlickGraphicUtil.drawTextCentered(g, font, x, y, "SNAXE");
    y += font.getLineHeight() * 2;

    g.setColor(Color.white);

    font = resources.getFont("FONT_NOKIA24");
    y += font.getLineHeight();
    if (curTime % 900 < 500) {
      SlickGraphicUtil.drawTextCentered(g, font, x, y, "INSERT COIN");
    }

    y += font.getLineHeight() * 4;

    SlickGraphicUtil.drawTextCentered(g, font, x, y, "Connect to wi-fi \"Snaxe\" and visit http://game/");
    y += font.getLineHeight();
    SlickGraphicUtil.drawTextCentered(g, font, x, y, "See instructions at bar for further info");
    y += font.getLineHeight();
    y += font.getLineHeight();

    int numQueued = 0;
    for (Player player : ((SnaxeGame) game).getSortedPlayers()) {
      if (player.getQueueJoinTime() > 0) {
        numQueued++;
      }
    }

    if (numQueued > 0) {
      SlickGraphicUtil.drawTextCentered(g, font, x, y, "Players waiting:");

      y += font.getLineHeight();

//      font = resources.getFont("FONT_NOKIA16");

      StringBuilder sb = new StringBuilder();
      int numThisLine = 0;

      for (Player player : ((SnaxeGame) game).getSortedPlayers()) {
        if (player.getQueueJoinTime() > 0) {
          if (numThisLine > 0) {
            sb.append(", ");
          }
          sb.append(player.getName());
          numThisLine++;

          if (numThisLine == QUEUED_PLAYERS_PER_LINE) {
            SlickGraphicUtil.drawTextCentered(g, font, x, y, sb.toString());
            y += font.getLineHeight();
            numThisLine = 0;
            sb = new StringBuilder();
          }
        }
      }

      if (numThisLine > 0) {
        SlickGraphicUtil.drawTextCentered(g, font, x, y, sb.toString());
        y += font.getLineHeight();
      }

      y += font.getLineHeight();

      font = resources.getFont("FONT_NOKIA24");
      if (subState == SubState.COUNTDOWN) {
        g.setColor(Color.red);
        SlickGraphicUtil.drawTextCentered(g, font, x, y, "NEXT GAME STARTS IN : " + (countdownLength - (curTime - countdownStartTime)) / 1000);
      }
    }

    font = resources.getFont("FONT_NOKIA16");
    g.setColor(Color.white);
    y = 700;
    SlickGraphicUtil.drawTextCentered(g, font, x, y, "(c) 1984 David C A Croft");
    y += font.getLineHeight();
    SlickGraphicUtil.drawTextCentered(g, font, x, y, "Edward Mccaughan, Katia Bonini");
  }

  public void update(GameContainer container, StateBasedGame game, int delta) throws SlickException
  {
    curTime += delta;

    if (forceStart) {
      startGame(game);
      return;
    }

    SnaxeGame snaxeGame = (SnaxeGame) game;

    int numPlayersQueued = 0;
    for (Player player : snaxeGame.getSortedPlayers()) {
      if (player.getQueueJoinTime() != 0) {
        numPlayersQueued++;
      }
    }

    if (subState == SubState.WAITING && numPlayersQueued >= SnaxeGame.MIN_PLAYERS) {
      // We now have the minimum #players, so can move into countdown
      subState = SubState.COUNTDOWN;
      countdownStartTime = curTime;
    }
    else if (subState == SubState.COUNTDOWN && numPlayersQueued < SnaxeGame.MIN_PLAYERS) {
      // We were in countdown, but no longer have the minimum number of players
      subState = SubState.WAITING;
      lastNotifiedCountdown = -1;

      // Any player left needs to be notified of 'queued' rather than 'prepare' state
      for (Player player : snaxeGame.getSortedPlayers()) {
        if (player.getQueueJoinTime() > 0) {
          try {
            player.sendMessage(Protocol.MESSAGE_OUT_QUEUED);
          }
          catch (IOException e) {
            log.log(Level.WARNING, "Unable to send prepare message to " + player);
          }
        }
      }
    }

    if (subState == SubState.COUNTDOWN) {
      // If we're in countdown, calculate countdown length depending on #players waiting
      int countdownPos = numPlayersQueued;
      if (countdownPos >= COUNTDOWN_LENGTHS.length) countdownPos = COUNTDOWN_LENGTHS.length - 1;
      countdownLength = (COUNTDOWN_LENGTHS[countdownPos] * 1000) + 999;

      // Countdown expired
      if (curTime >= countdownStartTime + countdownLength) {
        startGame(game);
        return;
      }

      // Inform any queued players that are in the next board of their countdown
      int countdownNow = (int) ((countdownLength - (curTime - countdownStartTime)) / 1000);
      if (countdownNow != lastNotifiedCountdown) {
        int playerNum = 0;
        for (Player player : snaxeGame.getSortedPlayers()) {
          if (player.getQueueJoinTime() > 0) {
            try {
              player.sendMessage(Protocol.MESSAGE_OUT_PREPARE + countdownNow);
            }
            catch (IOException e) {
              log.log(Level.WARNING, "Unable to send prepare message to " + player);
            }
          }
          playerNum++;
          if (playerNum == SnaxeGame.MAX_PLAYERS) break;
        }
        lastNotifiedCountdown = countdownNow;
      }
    }
  }

  private void startGame(StateBasedGame game)
  {
    /*GameState target = board.getState(PlayState.STATE_ID);

  final long start = System.currentTimeMillis();
  CrossStateTransition t = new CrossStateTransition(target)
  {
    public boolean isComplete()
    {
      return (System.currentTimeMillis() - start) > 2000;
    }

    public void init(GameState firstState, GameState secondState)
    {
    }
  };*/

    game.enterState(PlayState.STATE_ID); //, t, new EmptyTransition());

    subState = SubState.LAUNCHED;
  }

  public void forceStart()
  {
    forceStart = true;
  }
}
